package sorting;

public class HeapSort implements SortingAlgorithm<Integer>
{
    /**
     * Standard heapsort.
     * @param a an array of Comparable items.
     * @return sorting statistics.
     */
    @Override
    public Stats sort(Integer[] a) 
    {
        long moves = 0;
        long compares = 0;
        
        // Build heap.
        for (int i = a.length/2 - 1; i >= 0; i--) {
            Stats stats = percDown(a, i, a.length);
            moves += stats.moves;
            compares += stats.compares;
        }
        
        // Delete max.
        for (int i = a.length - 1; i > 0; i--) {
            Utilities.swapReferences(a, 0, i);
            moves += 2;
            Stats stats = percDown(a, 0, i);
            moves += stats.moves;
            compares += stats.compares;
        }

        return new Stats(moves, compares);
    }

    /**
     * Internal method for heapsort that is used in deleteMax and buildHeap.
     * @param a an array of Comparable items.
     * @index i the position from which to percolate down.
     * @int n the logical size of the binary heap.
     * @return sorting statistics.
     */
    private Stats percDown(Integer[] a, int i, int n) 
    {
        int child;
        int iSave = i;
        long moves = 0;
        long compares = 0;
        Integer tmp;
        
        for (tmp = a[i]; leftChild(i) < n; i = child) {
            child = leftChild(i);
            
            if ((child != n-1) && (a[child].compareTo(a[child+1]) < 0)) {
                child++;
                compares++;
            }
            
            if (child != n-1) compares++;
            
            compares++;
            if (tmp.compareTo(a[child]) < 0) {
                a[i] = a[child];
                moves++;
            }
            else break;
        }
        
        if (i != iSave) {
            a[i] = tmp;
            moves++;
        }
        
        return new Stats(moves, compares);
    }

    /**
     * Internal method for heapsort.
     * @param i the index of an item in the heap.
     * @return the index of the left child.
     */
    private int leftChild(int i) 
    {
        return 2*i + 1;
    }

    @Override
    public Stats sort(MyLinkedList<Integer> list) { return null; };
}
