package sorting;

public class MergeSortLinkedList implements SortingAlgorithm<Integer>
{
    /**
     * Mergesort algorithm for a list.
     * @param list a linked list of Comparable items to sort.
     * @return sorting statistics.
     */
    @Override
    public Stats sort(MyLinkedList<Integer> list) 
    {
        Stats stats = mergeSort(list);
        //System.out.print("list: "); printList(list);
        
        return stats;
    }
    
    /**
     * Internal method for mergesort that makes recursive calls.
     * @param list a linked list of Comparable items.
     * @return sorting statistics.
     */
    private Stats mergeSort(MyLinkedList<Integer> list)
    {
        Stats stats1 = new Stats();
        Stats stats2 = new Stats();
        Stats stats3 = new Stats();
        int moves = 0;
        
        if (list.size() > 1) {
            // Split the list roughly in half.
            MyLinkedList<Integer> lists[] = list.split();
            moves += 2;
            
            // Sort each sublist and merge.
            stats1 = mergeSort(lists[0]);
            stats2 = mergeSort(lists[1]);
            stats3 = merge(list, lists[0], lists[1]);
        }
        
        return new Stats(moves + stats1.moves + stats2.moves + stats3.moves,
                         stats1.compares + stats2.compares + stats3.compares);
    }
    
    /**
     * Internal method for mergesort that merges two sorted sublists.
     * @param list the merged sublists.
     * @param list1 the first sublist.
     * @param list2 the second sublist.
     * @return sorting statistics.
     */
    private Stats merge(MyLinkedList<Integer> list,
                        MyLinkedList<Integer> list1, 
                        MyLinkedList<Integer> list2)
    {
        MyNode<Integer> node1 = list1.head();
        MyNode<Integer> node2 = list2.head();
        long moves = 0;
        long compares = 0;
        
        list.empty();
        
        // Choose which node from sublist to add to the merged sublist.
        while((node1 != null) && (node2 != null)) {
            if (node1.data.compareTo(node2.data) <= 0) {
                MyNode<Integer> nextNode = node1.next;
                list.add(node1);
                node1 = nextNode;
            }
            else {
                MyNode<Integer> nextNode = node2.next;
                list.add(node2);
                node2 = nextNode;
            }
            moves++;
            compares++;
        }
        
        // Concatenate the rest of the first sublist to list.
        if (node1 != null) {
            list.concatenate(node1, list1);
            moves++;
        }
        
        // Concatenate the rest of the second sublist to list.
        if (node2 != null) {
            list.concatenate(node2, list2);
            moves++;
        }
        
        return new Stats(moves, compares);
    }
    
    @Override
    public Stats sort(Integer a[]) { return null; }
}
