package sorting;

/**
 * A linked list class to mergesort.
 * 
 * @author Ronald Mak
 *
 * @param <AnyType> the data type.
 */
public class MyLinkedList<AnyType> 
{
    private MyNode<AnyType> head;
    private MyNode<AnyType> tail;
    private int size = 0;
    
    /**
     * Constructor.
     */
    public MyLinkedList()
    {
        this(null, null, 0);
    }
    
    /**
     * Constructor. Create a linked list from a given linked list.
     * @param head the head of the given linked list.
     * @param tail the tail of the given linked list.
     * @param size the size of the given linked list.
     */
    public MyLinkedList(MyNode<AnyType> head, MyNode<AnyType> tail, int size)
    {
        this.head = head;
        this.head.prev = null;
        
        this.tail = tail;
        this.tail.next = null;
        
        this.size = size;
    }
    
    /**
     * Create a linked list from the data in an array.
     * @param a the array of data.
     */
    public MyLinkedList(AnyType a[])
    {
        for (AnyType data : a) add(new MyNode<AnyType>(data));
    }
    
    public MyNode<AnyType> head() { return head; }
    public MyNode<AnyType> tail() { return tail; }
    public int size()             { return size; }
    
    /**
     * Logically empty the linked list.
     */
    public void empty()
    {
        head = tail = null;
        size = 0;
    }
    
    /**
     * Add a node to the end of the linked list.
     * @param node the node to add.
     */
    public void add(MyNode<AnyType> node)
    {
        if (size == 0) {
            this.head = this.tail = node;
            node.prev = node.next = null;
        }
        else {
            this.tail.next = node;
            node.prev = this.tail;
            node.next = null;
            this.tail = node;
        }
        
        size++;
    }
    
    /**
     * Split a linked list roughly in half.
     * @return an array of the two sublists.
     */
    public MyLinkedList<AnyType>[] split()
    {
        int halfLength1 = size/2;
        int halfLength2 = size%2 == 0 ? halfLength1 : halfLength1 + 1;
        MyLinkedList<AnyType>[] lists = 
                (MyLinkedList<AnyType>[]) new MyLinkedList[2];
        
        // Get to the node at the midpoint.
        MyNode<AnyType> mid = head;
        for (int i = 1; i < halfLength1; i++) mid = mid.next;
        MyNode<AnyType> midNext = mid.next;
        
        // Create the two sublists.
        lists[0] = new MyLinkedList<AnyType>(this.head, mid, halfLength1);
        lists[1] = new MyLinkedList<AnyType>(midNext, this.tail, halfLength2);
        
        return lists;
    }
    
    /**
     * Concatenate another list to the end of this one,
     * starting at a given node of the other list.
     * @param otherNode the given node of the other list.
     * @param otherList the other list.
     */
    public void concatenate(MyNode<AnyType> otherNode, 
                            MyLinkedList<AnyType> otherList)
    {
        // Splice the other list to the end of this list.
        if (otherNode != null) {
            this.tail.next = otherNode;
            otherNode.prev = this.tail;
            this.tail = otherList.tail;
        }
        
        // Update this list's size.
        do {
            size++;
            otherNode = otherNode.next;
        } while (otherNode != null);
    }
}
