package sorting;

public class QuickSort implements SortingAlgorithm<Integer>
{
    private PivotStrategy pivotStrategy;
    
    public QuickSort(PivotStrategy pivotStrategy)
    {
        this.pivotStrategy = pivotStrategy;
    }
    
    /**
     * Quicksort algorithm.
     * @param a an array of Comparable items to sort.
     * @return sorting statistics.
     */
    @Override
    public Stats sort(Integer[] a) 
    {
        return quicksort(a, 0, a.length - 1);
    }
    
    /**
     * Internal quicksort method that makes recursive calls.
     * Uses the chosen pivot strategy.
     * @param a an array of Comparable items.
     * @param left the left-most index of the subarray.
     * @param right the right-most index of the subarray.
     * @return sorting statistics.
     */
    private Stats quicksort(Integer[] a, int left, int right) 
    {
        Stats stats = new Stats();
        
        if (left <= right) {
            
            Integer pivot = pivotStrategy.choosePivot(a, left, right, stats);
            int p = partition(a, left, right, pivot, stats);

            Stats stats1 = quicksort(a, left, p-1);  // Sort small elements
            Stats stats2 = quicksort(a, p+1, right); // Sort large elements
            
            stats.moves    += (stats1.moves    + stats2.moves);
            stats.compares += (stats1.compares + stats2.compares);
        } 
            
        return stats;
    }
    
    /**
     * Partition a portion of an array.
     * @param a the array to partition
     * @param left the first index of the portion to be partitioned
     * @param right the last index of the portion to be partitioned
     * @param pivot the pivot value
     * @param stats sorting statistics
     * @return the index of the pivot
     */
private int partition(Integer[] a, int left, int right, 
                      Integer pivot, Stats stats)
{
    int i = left-1;
    int j = right;
    
    while (i < j) {
    
        // Move i to the right.
        do {
            i++;
            stats.compares++;
        } while ((i <= right) && a[i].compareTo(pivot) < 0);
        
        // Move j to the left.
        do {
            j--;
            stats.compares++;
        } while ((j >= left)  && a[j].compareTo(pivot) > 0);
        
        if (i < j) {
            Utilities.swapReferences(a, i, j);
            stats.moves += 2;
        }
    }

    Utilities.swapReferences(a, i, right); // Restore pivot
    stats.moves += 2;
    
    return i;
}

    @Override
    public Stats sort(MyLinkedList<Integer> list) { return null; };
}
